from enum import Enum
from typing import Generic, TypeVar, Type
from asyncio import Queue
from pydantic.generics import GenericModel
from pydantic import BaseModel

P = TypeVar('P')
A = TypeVar('A', bound=Enum)


class PayloadBase(BaseModel):
    name: str


class ErrorPayload(PayloadBase):
    error: Exception

    class Config:
        arbitrary_types_allowed = True


class UIEvent(GenericModel, Generic[A, P]):
    action: A
    payload: P


T = TypeVar('T', bound=UIEvent)


class UIBase(Generic[T]):
    def __init__(self, q: Queue):
        self._q = q

    async def __call__(self):
        while True:
            item: T = await self._q.get()
            handler = getattr(self, f"on_{item.action.value}")
            handler(item.payload)
            self._q.task_done()

    def __enter__(self):
        return self

    def __exit__(self):
        return None



from . import ArgsBase, BaseCommand

from ..config import ConfigItem
from typing import List, Union
from doter.ui.install_ui import InstallUI, InstallUIEvent, InstallAction
from ..utils import sh
from os import symlink

from ..utils import LoggerMixin
from shutil import rmtree
from pydantic import Field


class InstallArgs(ArgsBase):
    links_only: bool = Field(False,
                             description="only create symlinks without hooks",
                             alias="l")
    dry_run: bool = Field(False, description="Show execution plan", alias="d")
    force: bool = Field(False, description="Override existing files/links")


class InstallCommand(BaseCommand[InstallArgs, InstallUIEvent], LoggerMixin):
    trigger = "install"
    UIClass = InstallUI

    def create_link(self, key: str, managed_path: str, link_path: str):
        managed_p = self.resolve_path(managed_path)
        link_p = self.resolve_path(link_path)

        if link_p.exists() and not self.args.force:
            raise FileExistsError(
                f"target path '{link_path}' already exists and --force is not specified "
            )

        self._dispatch_ui_action(
            InstallAction.STEP, {
                "name": key,
                "msg": f"Linking {managed_path} -> {link_path}"
            })

        rmtree(link_p, ignore_errors=True)
        link_p.parent.mkdir(parents=True, exist_ok=True)
        symlink(managed_p, link_p)

    async def run(self, key: str, item: ConfigItem):

        try:
            total_steps = self.calculate_steps(item)
            self._dispatch_ui_action(InstallAction.INIT, {
                "name": key,
                "total_steps": total_steps
            })
            # Execute Pre-install hooks
            await self.execute_hooks(key, item.before)
            # Create Symbolic links
            for managed_path, original_path in item.mappings.items():
                self.create_link(key, managed_path, original_path)
            # Execute Post-install hooks
            await self.execute_hooks(key, item.after)
            self._dispatch_ui_action(InstallAction.DONE, {"name": key})
        # Stop current task when exceptions are raised
        except Exception as e:
            self.error(e)
            self._dispatch_ui_action(InstallAction.ERROR, {
                "name": key,
                "error": e,
            })
            return

    def calculate_steps(self, item: ConfigItem):
        res = 0 if item.before is None or self.args.links_only else len(
            item.before)
        res += 0 if item.after is None or self.args.links_only else len(
            item.after)
        res += len(item.mappings)
        return res

    async def execute_hooks(self, key: str, hooks: Union[List[str], None]):
        if hooks is None:
            return

        for hook in hooks:
            self._dispatch_ui_action(InstallAction.STEP, {
                "name": key,
                "msg": f"executing: {hook}"
            })
            await sh(hook)

    def _dispatch_ui_action(self, action: InstallAction, payload: dict):
        self.notify_ui(InstallUIEvent(action=action, payload=payload))



from asyncio import subprocess
from typing import Any, Callable, Generic, List, Optional, TypeVar
import logging
from doter.exceptions import SubprocessError
import inspect


def copysig(from_func, *args_to_remove):
    def wrap(func):
        #add and remove parameters
        oldsig = inspect.signature(from_func)
        oldsig = _remove_args(oldsig, args_to_remove)
        newsig = _add_args(oldsig, func)

        #write some code for a function that we can exec
        #the function will have the correct signature and forward its arguments to the real function
        code = '''
def {name}{signature}:
    {func}({args})
'''.format(name=func.__name__,
           signature=newsig,
           func='_' + func.__name__,
           args=_forward_args(oldsig, newsig))
        globs = {'_' + func.__name__: func}
        exec(code, globs)
        newfunc = globs[func.__name__]

        #copy as many attributes as possible
        newfunc.__doc__ = func.__doc__
        newfunc.__module__ = func.__module__
        #~ newfunc.__closure__= func.__closure__
        #~ newfunc.__code__.co_filename= func.__code__.co_filename
        #~ newfunc.__code__.co_firstlineno= func.__code__.co_firstlineno
        return newfunc

    return wrap


def _collectargs(sig):
    """
    Writes code that gathers all parameters into "self" (if present), "args" and "kwargs"
    """
    arglist = list(sig.parameters.values())

    #check if the first parameter is "self"
    selfarg = ''
    if arglist:
        arg = arglist[0]
        if arg.name == 'self':
            selfarg = 'self, '
            del arglist[0]

    #all named parameters will be passed as kwargs. args is only used for varargs.
    args = 'tuple(), '
    kwargs = ''
    kwarg = ''
    for arg in arglist:
        if arg.kind in (arg.POSITIONAL_ONLY, arg.POSITIONAL_OR_KEYWORD,
                        arg.KEYWORD_ONLY):
            kwargs += '("{0}",{0}), '.format(arg.name)
        elif arg.kind == arg.VAR_POSITIONAL:
            #~ assert not args
            args = arg.name + ', '
        elif arg.kind == arg.VAR_KEYWORD:
            assert not kwarg
            kwarg = 'list({}.items())+'.format(arg.name)
        else:
            assert False, arg.kind
    kwargs = 'dict({}[{}])'.format(kwarg, kwargs[:-2])

    return '{}{}{}'.format(selfarg, args, kwargs)


def _forward_args(args_to_collect, sig):
    collect = _collectargs(args_to_collect)

    collected = {arg.name for arg in args_to_collect.parameters.values()}
    args = ''
    for arg in sig.parameters.values():
        if arg.name in collected:
            continue

        if arg.kind == arg.VAR_POSITIONAL:
            args += '*{}, '.format(arg.name)
        elif arg.kind == arg.VAR_KEYWORD:
            args += '**{}, '.format(arg.name)
        else:
            args += '{0}={0}, '.format(arg.name)
    args = args[:-2]

    code = '{}, {}'.format(collect, args) if args else collect
    return code


def _remove_args(signature, args_to_remove):
    """
    Removes named parameters from a signature.
    """
    args_to_remove = set(args_to_remove)
    varargs_removed = False
    args = []
    for arg in signature.parameters.values():
        if arg.name in args_to_remove:
            if arg.kind == arg.VAR_POSITIONAL:
                varargs_removed = True
            continue

        if varargs_removed and arg.kind == arg.KEYWORD_ONLY:  #if varargs have been removed, there are no more keyword-only parameters
            arg = arg.replace(kind=arg.POSITIONAL_OR_KEYWORD)

        args.append(arg)

    return signature.replace(parameters=args)


def _add_args(sig, func):
    """
    Merges a signature and a function into a signature that accepts ALL the parameters.
    """
    funcsig = inspect.signature(func)

    #find out where we want to insert the new parameters
    #parameters with a default value will be inserted before *args (if any)
    #if parameters with a default value exist, parameters with no default value will be inserted as keyword-only AFTER *args
    vararg = None
    kwarg = None
    insert_index_default = None
    insert_index_nodefault = None
    default_found = False
    args = list(sig.parameters.values())
    for index, arg in enumerate(args):
        if arg.kind == arg.VAR_POSITIONAL:
            vararg = arg
            insert_index_default = index
            if default_found:
                insert_index_nodefault = index + 1
            else:
                insert_index_nodefault = index
        elif arg.kind == arg.VAR_KEYWORD:
            kwarg = arg
            if insert_index_default is None:
                insert_index_default = insert_index_nodefault = index
        else:
            if arg.default != arg.empty:
                default_found = True

    if insert_index_default is None:
        insert_index_default = insert_index_nodefault = len(args)

    #find the new parameters
    #skip the first two parameters (args and kwargs)
    newargs = list(funcsig.parameters.values())
    if not newargs:
        raise Exception(
            'The decorated function must accept at least 2 parameters')
    #if the first parameter is called "self", ignore the first 3 parameters
    if newargs[0].name == 'self':
        del newargs[0]
    if len(newargs) < 2:
        raise Exception(
            'The decorated function must accept at least 2 parameters')
    newargs = newargs[2:]

    #add the new parameters
    if newargs:
        new_vararg = None
        for arg in newargs:
            if arg.kind == arg.VAR_POSITIONAL:
                if vararg is None:
                    new_vararg = arg
                else:
                    raise Exception(
                        'Cannot add varargs to a function that already has varargs'
                    )
            elif arg.kind == arg.VAR_KEYWORD:
                if kwarg is None:
                    args.append(arg)
                else:
                    raise Exception(
                        'Cannot add kwargs to a function that already has kwargs'
                    )
            else:
                #we can insert it as a positional parameter if it has a default value OR no other parameter has a default value
                if arg.default != arg.empty or not default_found:
                    #do NOT change the parameter kind here. Leave it as it was, so that the order of varargs and keyword-only parameters is preserved.
                    args.insert(insert_index_default, arg)
                    insert_index_nodefault += 1
                    insert_index_default += 1
                else:
                    arg = arg.replace(kind=arg.KEYWORD_ONLY)
                    args.insert(insert_index_nodefault, arg)
                    if insert_index_default == insert_index_nodefault:
                        insert_index_default += 1
                    insert_index_nodefault += 1

        #if varargs need to be added, insert them before keyword-only arguments
        if new_vararg is not None:
            for i, arg in enumerate(args):
                if arg.kind not in (arg.POSITIONAL_ONLY,
                                    arg.POSITIONAL_OR_KEYWORD):
                    break
            else:
                i += 1
            args.insert(i, new_vararg)

    return inspect.Signature(args, return_annotation=funcsig.return_annotation)


F = TypeVar('F', bound=Callable[..., Any])


class LoggerMixin(object):
    def __init__(self):
        self._logger = logging.getLogger(__file__)

    def info(self, *args, **kwargs):
        self._logger.info(*args, **kwargs)

    def error(self, *args, **kwargs):
        self._logger.error(*args, **kwargs)

    def warn(self, *args, **kwargs):
        self._logger.warn(*args, **kwargs)

    def debug(self, *args, **kwargs):
        self._logger.debug(*args, **kwargs)


async def sh(cmd: str):
    proc = await subprocess.create_subprocess_shell(
        cmd,
        stderr=subprocess.PIPE,
        shell=True,
    )
    _, stderr = await proc.communicate()
    retcode = proc.returncode
    if retcode != 0:
        raise SubprocessError(cmd, retcode, stderr)



from pydantic import BaseModel
from os import PathLike
from typing import Union, List, Dict


class ConfigItem(BaseModel):
    mappings: Dict[str, str]
    before: Union[List[str], None]
    after: Union[List[str], None]


class ConfigFile(BaseModel):
    version: int
    items: Dict[str, ConfigItem]

    def to_file(self, path: PathLike):
        self.dict()
        return FileExistsError()

    @classmethod
    def from_file(cls, path: PathLike):
        from yaml import safe_load
        f = open(path, 'r')
        values = safe_load(f)
        f.close()
        return ConfigFile(**values)



class SubprocessError(Exception):
    def __init__(self, cmd: str, retcode: int, msg: str) -> None:
        self._retcode = retcode
        self._err_msg = msg
        self._cmd = cmd
        super().__init__(self.msg())

    def msg(self):
        return f"cmd '{self._cmd}' Exited with code {self._retcode}"

    @property
    def retcode(self):
        return self._retcode

    @property
    def err_msg(self):
        return self.msg

    @property
    def cmd(self):
        return self._cmd


class FileAlreadExistsError(Exception):
    def __init__(self, path: str):
        self._path = path

    def msg(self):
        return f"file '{self._path}' already "


import os
from os import PathLike
from doter.config import ConfigItem

from doter.utils import LoggerMixin
from . import ArgsBase, BaseCommand
from doter.ui.cleanup_ui import CleanupAction, CleanupUI, CleanupUIEvent
from shutil import rmtree


class CleanupArgs(ArgsBase):
    force: bool = False,
    dry_run: bool = False,


class CleanupCommand(BaseCommand[CleanupArgs, CleanupUIEvent], LoggerMixin):
    trigger = "clean"
    UIClass = CleanupUI

    def _remove_link(self, p: str, name: str):
        path = self.resolve_path(p)
        if not path.is_symlink() and not self.args.force:
            raise ItemIsNotSymlinkError(path)

        if path.is_symlink():
            path.unlink()
        elif self.args.force:
            if path.is_dir():
                rmtree(path)
            else:
                os.remove(path)

        self._dispatch_ui_action(CleanupAction.DONE, {"name": name})
        self.info(f"Removed '{p}'")

    def _dispatch_ui_action(self, action: CleanupAction, payload: dict):
        self.notify_ui((CleanupUIEvent(action=action, payload=payload)))

    async def run(self, key: str, config: ConfigItem):
        try:
            for link_path in config.mappings.values():
                self._remove_link(link_path, key)
        except ItemIsNotSymlinkError as e:
            self._dispatch_ui_action(CleanupAction.ERROR, {
                "name": key,
                "error": e
            })


class ItemIsNotSymlinkError(Exception):
    def __init__(self, path: PathLike) -> None:
        self._path = path
        super().__init__()

    def __str__(self):
        return f"Item: '{self._path} is not a symlink and '--force' is not set"



from asyncio import Queue
from rich.progress import Progress, SpinnerColumn, TextColumn, ProgressColumn, Task
from pydantic import BaseModel
from rich.text import Text
from . import UIBase, UIEvent, PayloadBase, ErrorPayload
from typing import Dict, Optional, Union, Type, Any
from enum import Enum


class InstallAction(str, Enum):
    INIT = "init"
    STEP = "step"
    ERROR = "error"
    DONE = "done"


class InitPayload(PayloadBase):
    total_steps: int


class StepPayload(PayloadBase):
    msg: str


InstallUIPayload = Union[InitPayload, StepPayload, ErrorPayload, PayloadBase]
InstallUIEvent = UIEvent[InstallAction, InstallUIPayload]


class InstallUI(UIBase[InstallUIPayload]):
    def __init__(self, queue: Queue):
        self._progress = Progress(
            SpinnerColumn(),
            CurrentProgressColumn(),
            TextColumn("{task.description}"),
        )
        self._name2id: Dict[str, int] = {}
        super().__init__(queue)

    def on_init(self, payload: InitPayload):
        task_id = self._progress.add_task(
            f"{payload.name}: Preparing",
            total=payload.total_steps,
        )
        self._name2id[payload.name] = task_id

    def on_step(self, payload: StepPayload):
        name = payload.name
        task_id = self._name2id[name]
        self._progress.update(task_id,
                              advance=1,
                              description=f"{payload.name}: f{payload.msg}")

    def on_done(self, payload: PayloadBase):
        task_id = self.get_task_id(payload.name)
        self._progress.update(task_id,
                              completed=True,
                              advance=0,
                              description=f"{payload.name}: Done")

    def on_error(self, payload: ErrorPayload):
        task_id = self.get_task_id(payload.name)
        self._progress.update(task_id,
                              completed=True,
                              description=f"{payload.name}: {payload.error}")
        self._progress.print(payload.error)

    def get_task_id(self, name: str):
        return self._name2id[name]

    def __enter__(self):
        self._progress.__enter__()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[Any],
    ):
        self._progress.__exit__(exc_type, exc_val, exc_tb)


class CurrentProgressColumn(ProgressColumn):
    def render(self, task: "Task"):
        text = ""
        if not task.finished:
            text = f"{task.completed}/{task.total}"
            text = Text(text)
        else:
            text = f"{task.total}/{task.total}"
        return text


from enum import Enum
from . import UIBase, UIEvent, PayloadBase, ErrorPayload
from typing import Union
from rich.console import Console


class CleanupAction(str, Enum):
    DONE = "done"
    ERROR = "error"


CleanupUIPayload = Union[PayloadBase, ErrorPayload]
CleanupUIEvent = UIEvent[CleanupAction, CleanupUIPayload]


class CleanupUI(UIBase[CleanupUIPayload]):
    def __init__(self):
        self._console = Console()

    def on_done(self, payload: PayloadBase):
        item = payload.name
        self._console.print(f"Done removing: `{item}`")

    def on_error(self, payload: ErrorPayload):
        err = payload.error
        self._console.print_exception(err)



from argparse import ArgumentParser

from typing import List, Generic, Type, TypeVar, get_args, Dict, Any

from ..ui import UIBase, UIEvent
from pydantic import BaseModel
from pydantic.fields import SHAPE_LIST, SHAPE_TUPLE, SHAPE_ITERABLE, SHAPE_SET, SHAPE_DEQUE, SHAPE_SEQUENCE

from ..config import ConfigFile, ConfigItem
from typing import List
from asyncio import Queue, gather
from logging import getLogger

ITERABLE_SHAPES = [
    SHAPE_LIST, SHAPE_TUPLE, SHAPE_ITERABLE, SHAPE_SET, SHAPE_DEQUE,
    SHAPE_SEQUENCE
]
from pathlib import Path


class ArgsBase(BaseModel):
    items: List[str] = []
    config: str = "~/.dotfiles/config.yml"


Args = TypeVar('Args', bound=ArgsBase)

Event = TypeVar('Event', bound=UIEvent)


class BaseCommand(Generic[Args, Event]):
    trigger = ""
    UIClass = UIBase

    @classmethod
    def from_args_dict(cls, args_dict: Dict[str, Any]):
        args = cls.arg_class(**args_dict)
        return cls(args)

    def __init__(self, args: Args):
        self.args = args
        self._queue = Queue()
        self._logger = getLogger(__name__)

    async def run(self, key: str, config: ConfigItem):
        pass

    async def __call__(self, config: ConfigFile):
        keys = self.args.items if len(
            self.args.items) != 0 else config.items.keys()
        await gather(*[self._prepare_run(key, config) for key in keys])

    async def _prepare_run(self, key: str, config: ConfigFile):
        try:
            item = config.items[key]
            await self.run(key, item)
        except KeyError as e:
            print(f"{key} does not exist")

    @classmethod
    def create_parser_args(cls, parser: ArgumentParser):
        arg_class = cls.arg_class
        for key, field in arg_class.__fields__.items():
            params = {
                "default": field.default,
                "help": field.field_info.description or "",
            }
            short_alias = None
            if field.alias != field.name and isinstance(field.alias, str):
                if len(field.alias) == 1:
                    short_alias = f"-{field.alias}"
                else:
                    raise ValueError(
                        f"{field.alias} havs length {len(field.alias)}, expected to be 1"
                    )

            if isinstance(field.default, bool) and not field.default:
                params['action'] = "store_true"
            elif isinstance(field.default, bool) and field.default:
                params['action'] = "store_false"
            field_name = key if field.shape in ITERABLE_SHAPES else f"--{key}"

            if not field_name.startswith("--"):
                params["nargs"] = "?"
            if short_alias:
                parser.add_argument(field_name, short_alias, **params)
            else:
                parser.add_argument(field_name, **params)

    def resolve_path(self, p: str):
        path = Path(p).expanduser()
        if not path.is_absolute():
            return self.config_dir / path
        else:
            return path

    @property
    def config_dir(self):
        p = Path(self.args.config)
        return p.parent.absolute()

    @classmethod
    @property
    def arg_class(cls) -> Type[Args]:
        return get_args(cls.__orig_bases__[0])[0]

    @classmethod
    @property
    def ui_event_class(cls) -> Type[Event]:
        return get_args(cls.__orig_bases__[0])[1]

    @property
    def queue(self):
        return self._queue

    def notify_ui(self, event: Event):
        self._queue.put_nowait(event)
